import gym
import numpy
from numpy import random


class AnyMDPSolverOTS(object):
    """
    Implementing the RL Solver  of the paper
    Hu, Bingshan, et al. "Optimistic Thompson sampling-based algorithms for 
        episodic reinforcement learning." 
        Uncertainty in Artificial Intelligence. PMLR, 2023.
    """
    def __init__(self, env, gamma=0.99, c=0.01, alpha=0.10, max_steps=4000):
        """
        The constructor for the class AnyMDPSolverQ
        The exploration strategy is controlled by UCB-H with c as its hyperparameter. Increasing c will encourage exploration
        Simulation of the ideal policy when the ground truth is not known
        """
        if(not env.task_set):
            raise Exception("AnyMDPEnv is not initialized by 'set_task', must call set_task first")
        self.n_actions = env.action_space.n
        self.n_states = env.observation_space.n
       
        self.est_r = numpy.zeros((self.n_states, self.n_actions))
        self.est_r_cnt = numpy.zeros((self.n_states, self.n_actions))
        self.est_t = 0.1 * numpy.ones((self.n_states, self.n_actions, self.n_states))
        self.gamma = gamma
        self.lr = 1.0 - alpha
        self._c = c / (1.0 - self.gamma)
        self.max_steps = max_steps
        self.reset_states = numpy.zeros(self.n_states)
        for i,s in enumerate(env.state_mapping):
            self.reset_states[s] = env.reset_states[i]

        self.value_matrix = numpy.zeros((self.n_states, self.n_actions))
        self.est_r_global_avg = 0
        self.est_r_global_cnt = 0
        self.alpha = alpha


    def learner(self, s, a, ns, r, done):
        # Update the environment model estimation
        if(self.est_r_cnt[s,a] > 0):
            self.est_r[s,a] += (r - self.est_r[s,a])/self.est_r_cnt[s,a]
        else:
            self.est_r[s,a] = r

        if(self.est_r_global_cnt > 0):
            self.est_r_global_avg += (r - self.est_r_global_avg) * self.alpha
        else:
            self.est_r_global_avg = r

        self.est_r_cnt[s,a]+= 1
        self.est_t[s,a,ns] += 1
        self.est_r_global_cnt += 1

        est_t = self.est_t / numpy.sum(self.est_t, axis=-1, keepdims=True)

        if(not done):
            est_target = numpy.sum(est_t[s, a] * numpy.max(self.value_matrix, axis=-1))
            est_value = self.est_r[s,a] + self.gamma * est_target
        else:
            est_target = numpy.sum(self.reset_states * numpy.max(self.value_matrix, axis=-1))
            est_value = self.est_r[s,a] + self.gamma * est_target
        self.value_matrix[s,a] += self.alpha * (est_value - self.value_matrix[s,a])


    def policy(self, state):
        # Apply UCB with dynamic noise (Thompson Sampling)
        values = self._c * numpy.sqrt(numpy.log(self.max_steps + 1) / numpy.clip(self.est_r_cnt[state], 1.0, None)) * \
                numpy.maximum(numpy.random.randn(self.n_actions), 0) + \
                self.value_matrix[state]
        return numpy.argmax(values)


#   Copyright (c) 2021 DeepEvolution Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file is used to generate data for meta language models

import sys
import argparse
import random
import time
import _io
import numpy
import pickle
from l3c.metalang import MetaLangV1
from l3c.metalang import MetaLangV2
from l3c.metalang import TaskSamplerV1, TaskSamplerV2

def _text_io(tokens, output_stream):
    if(isinstance(output_stream, _io.TextIOWrapper)):
        need_close = False
    elif(isinstance(output_stream, str)):
        output_stream = open(output_stream, "w")
        need_close = True
    else:
        output_stream = sys.stdout
        need_close = False
    for i in range(tokens.shape[0]):
        output_stream.write("\t".join(map(str, tokens[i].tolist())))
        output_stream.write("\n")
    if(need_close):
        output_stream.close()

def metalang_generator(sample_type='sequences',
                       version='v1',
                       vocab_size=64,
                       patterns_number=10,
                       n_gram=3,
                       error_rate=0.15,
                       embedding_size=16,
                       hidden_size=64,
                       samples=1000,
                       lambda_weight=5.0,
                       batch_size=1,
                       task_file=None,
                       sequence_length=4096,
                       output_type='txt',
                       output=None
                       ):
    seed_base = int(time.time()*1000 % 1000000)

    if(sample_type == 'tasks'):
        if(output is None):
            raise Exception("Must specify --output when sample_type is tasks")
        if(version=='v1'):
            tasks = [TaskSamplerV1(seed=i+seed_base, 
                                   n_vocab=vocab_size, 
                                   n_patterns=patterns_number, 
                                   n_gram=n_gram, 
                                   error_ratio=error_rate) for i in range(samples)]
        elif(version=='v2'):
            tasks = [TaskSamplerV2(seed=i+seed_base,
                                   n_vocab=vocab_size,
                                   n_emb=embedding_size,
                                   n_hidden=hidden_size,
                                   n_gram=n_gram,
                                   _lambda=lambda_weight
                                   ) for i in range(samples)]
        output_file_name = output
        if(not output_file_name.endswith('.pkl')):
            output_file_name += '.pkl'
        pickle.dump(tasks, open(output_file_name, 'wb'))
    else:
        if(version == 'v1'):
            env = MetaLangV1(L=sequence_length)
        elif(version == 'v2'):
            env = MetaLangV2(L=sequence_length)

        if(task_file is None):
            tasks = None
        else:
            tasks = pickle.load(open(task_file, 'rb'))

        batch_size = batch_size
        if(tasks is None):
            # Generate a unique task for each sample
            batch_size = 1
            if(version=='v1'):
                tasks = [TaskSamplerV1(seed=i+seed_base, 
                                       n_vocab=vocab_size, 
                                       n_patterns=patterns_number, 
                                       n_gram=n_gram, 
                                       error_ratio=error_rate) for i in range(samples)]
            elif(version=='v2'):
                tasks = [TaskSamplerV2(seed=i+seed_base,
                                       n_vocab=vocab_size,
                                       n_emb=embedding_size,
                                       n_hidden=hidden_size,
                                       n_gram=n_gram,
                                       _lambda=lambda_weight
                                       ) for i in range(samples)]
        batch_number = (samples - 1) // batch_size + 1
        tokens = []
        if(len(tasks) < batch_number):
            tasks = tasks * ((batch_number - 1)//len(tasks) + 1)
        random.shuffle(tasks)
        for i in range(batch_number):
            env.set_task(tasks[i])
            seed_base = int(time.time()*1000 % 1000000)
            token = env.batch_generator(batch_size, seed=i+seed_base)
            tokens.append(token)
        tokens=numpy.concatenate(tokens, axis=0)

        if(output_type == 'npy'):
            numpy.save(output, tokens)
        elif(output_type == 'txt'):
            _text_io(tokens, output)


if __name__=='__main__':
    parser = argparse.ArgumentParser(description='Generating Meta Language Tasks or Sequences')
    parser.add_argument('--version', type=str, choices=['v1', 'v2'], default='v2')
    parser.add_argument('--sample_type', type=str, choices=['tasks', 'sequences'], default='sequences', help='Generate tasks or sequences')
    parser.add_argument('--task_file', type=str, default=None, help='Specify task file to generate from if the sample_type is sequences. Default will generate task on the fly.')
    parser.add_argument('--vocab_size', type=int, default=64)
    parser.add_argument('--embedding_size', type=int, default=16)
    parser.add_argument('--hidden_size', type=int, default=16)
    parser.add_argument('--patterns_number', type=int, default=10)
    parser.add_argument('--error_rate', type=float, default=0.20)
    parser.add_argument('--n_gram', nargs='+', type=int, default=[3,4,5,6], help="A [List of] length n used for generating tasks")
    parser.add_argument('--lambda_weight', type=float, default=5.0, help="Lambda weight multiplied for softmax sampling in MetaLangV2")
    parser.add_argument('--batch_size', type=int, default=1)
    parser.add_argument('--sequence_length', type=int, default=4096)
    parser.add_argument('--samples', type=int, default=100, help='number of sequences / tasks to generate')
    parser.add_argument('--output_type', type=str, choices=['txt', 'npy'], default='txt')
    parser.add_argument('--output', type=str, default=None)

    args = parser.parse_args()
    metalang_generator(**vars(args))



import gym
import sys
import argparse
import l3c.mazeworld
from l3c.mazeworld import MazeTaskSampler

if __name__=='__main__':
    parser = argparse.ArgumentParser(description='Playing the maze world demo with your keyboard')
    parser.add_argument('--max_steps', type=int, default=1000000)
    parser.add_argument('--visibility_3D', type=float, default=12, help="3D vision range, Only valid in 3D mode")
    parser.add_argument('--save_replay', type=str, default=None, help="Save the replay trajectory in file")

    args = parser.parse_args()

    # Create the environment
    maze_env = gym.make("mazeworld-v2", 
                        enable_render=True, 
                        max_steps=args.max_steps, 
                        visibility_3D=args.visibility_3D, 
                        command_in_observation=False,
                        render_scale=320)

    # Sample and set the task
    task = MazeTaskSampler()
    maze_env.set_task(task)

    maze_env.reset()
    done=False
    sum_reward = 0

    while not done:
        maze_env.render()
        state, reward, done, _ = maze_env.step(None)
        sum_reward += reward
        print("Instant r = %.2f, Accumulate r = %.2f" % (reward, sum_reward))
        if(maze_env.key_done):
            break
    print("Episode is over! You got %.2f score."%sum_reward)

    if(args.save_replay is not None):
        maze_env.save_trajectory(args.save_replay)



if __name__=="__main__":
    import gym
    import l3c.anymdp
    from l3c.anymdp import  AnyMDPSolverOpt, AnyMDPSolverOTS, AnyMDPSolverQ, AnyMDPTaskSampler

    task = AnyMDPTaskSampler(state_space=64, 
                             action_space=5,
                             min_state_space=32,
                             verbose=True)
    max_steps = 32000
    prt_freq = 1000

    # Test D2C Random Policy
    env = gym.make("anymdp-d2c-v0")
    env.set_task(task)

    # Test Random Policy
    state, info = env.reset()
    acc_reward = 0
    epoch_reward = 0
    done = False

    steps = 0
    while steps < max_steps:
        action = env.action_space.sample()
        state, reward, done, info = env.step(action)
        acc_reward += reward
        epoch_reward += reward
        steps += 1
        if(steps % prt_freq == 0 and steps > 0):
            print("Step:{}\tEpoch Reward: {}".format(steps, epoch_reward))
            epoch_reward = 0
        if(done):
            state, info = env.reset()
    print("D2C Random Policy Summary: {}".format(acc_reward))

    # Test Random Policy
    env = gym.make("anymdp-v0")
    env.set_task(task)
    state, info = env.reset()
    acc_reward = 0
    epoch_reward = 0
    done = False

    steps = 0
    while steps < max_steps:
        action = env.action_space.sample()
        state, reward, done, info = env.step(action)
        acc_reward += reward
        epoch_reward += reward
        steps += 1
        if(steps % prt_freq == 0 and steps > 0):
            print("Step:{}\tEpoch Reward: {}".format(steps, epoch_reward))
            epoch_reward = 0
        if(done):
            state, info = env.reset()
    print("Random Policy Summary: {}".format(acc_reward))

    # Test AnyMDPSolverOpt
    solver = AnyMDPSolverOpt(env)
    state, info = env.reset()
    done = False
    acc_reward = 0
    epoch_reward = 0
    steps = 0

    while steps < max_steps:
        action = solver.policy(state)
        state, reward, done, info = env.step(action)
        acc_reward += reward
        epoch_reward += reward
        steps += 1
        if(steps % prt_freq == 0 and steps > 0):
            print("Step:{}\tEpoch Reward: {}".format(steps, epoch_reward))
            epoch_reward = 0
        if(done):
            state, info = env.reset()
            state_list = []
    print("Optimal Solver Summary:  {}".format(acc_reward))

    # Test AnyMDPSolverQ
    solver = AnyMDPSolverQ(env)
    state, info = env.reset()
    done = False
    acc_reward = 0
    epoch_reward = 0
    steps = 0

    while steps < max_steps:
        action = solver.policy(state)
        next_state, reward, done, info = env.step(action)
        solver.learner(state, action, next_state, reward, done)
        acc_reward += reward
        epoch_reward += reward
        state = next_state
        steps += 1
        if(steps % prt_freq == 0 and steps > 0):
            print("Step:{}\tEpoch Reward: {}".format(steps, epoch_reward))
            epoch_reward = 0
        if(done):
            state, info = env.reset()
    print("Q Solver Summary: {}".format(acc_reward))

    # Test AnyMDPSolverOTS
    solver = AnyMDPSolverOTS(env)
    state, info = env.reset()
    done = False
    acc_reward = 0
    epoch_reward = 0
    steps = 0

    while steps < max_steps:
        action = solver.policy(state)
        next_state, reward, done, info = env.step(action)
        solver.learner(state, action, next_state, reward, done)
        acc_reward += reward
        epoch_reward += reward
        state = next_state
        steps += 1
        if(steps % prt_freq == 0 and steps > 0):
            print("Step:{}\tEpoch Reward: {}".format(steps, epoch_reward))
            epoch_reward = 0
        if(done):
            state, info = env.reset()
    print("OTS Solver Summary: {}".format(acc_reward))

    print("Test Passed")


"""
Any MDP Task Sampler
"""
import numpy
import gym
import pygame
import time
from numpy import random
from copy import deepcopy
from l3c.utils import pseudo_random_seed
from l3c.anymdp.solver import check_task_trans, check_task_rewards


def reward_sampler_sas(state_space:int, 
                   action_space:int, 
                   reward_sparsity:float=0.25,
                   exp_positive_ratio:float=0.10,
                   reward_noise_max:float=0.30):
    reward_mask = numpy.zeros((state_space,))
    reward_sparsity = min(reward_sparsity * random.exponential(1.0), 1.0)
    positive_ratio = min(positive_ratio * random.exponential(1.0), 1.0)
    reward_mask = random.binomial(1, reward_sparsity, size=(state_space,))
    reward_noise_mask = random.binomial(1, reward_sparsity, size=(state_space,))
    reward_noise = reward_noise_mask * reward_noise_max  * random.random() * random.rand(state_space)

    positive_ratio = min(exp_positive_ratio * random.exponential(1.0), 1.0)
    sign_mask = 2.0 * random.binomial(1, positive_ratio, size=(state_space,)) - 1
    reward_matrix = numpy.abs(random.normal(loc=0, scale=1.0, size=(state_space,))) * reward_mask * sign_mask
    return reward_matrix, reward_noise

def comb_sampler(n):
    comb_type = bin(random.randint(0, 2 ** n))[3:]
    if(len(comb_type) < n):
        comb_type = '0' * (n - len(comb_type)) + comb_type
    else:
        comb_type = comb_type[-n:]
    return comb_type

def reward_sampler(task:dict,
                   state_space:int, 
                   action_space:int,
                   reward_noise_choice:list):

    if("reset_triggers_positive" not in task):
        return {}

    reward_scale = random.uniform(0.01, 0.10) # Scale the reward noise
    reward_noise_type =  random.choice(reward_noise_choice)

    if(state_space < 2):
        reward_matrix = numpy.random.random((1, action_space, 1)) * reward_scale
        reward_noise = 0.3 * numpy.random.random((1, action_space, 1)) * reward_scale
        return {"reward": reward_matrix,
                "reward_noise": reward_noise,
                "reward_noise_type": reward_noise_type}

    # Sample Pitfalls
    v_scale = random.uniform(10, 100) * reward_scale
    rewards_neg = - numpy.random.random(size=[state_space]) * v_scale \
        * task["reset_triggers_negative"]
    # Sample Sucesses
    rewards_pos = task["reset_triggers_positive"] * \
        numpy.clip(
            numpy.random.exponential(scale=v_scale,
            size=state_space), v_scale, None)

    # Crucial Rewards
    rewards = rewards_neg + rewards_pos

    # Reshape the reward to ns, na, ns
    reward_matrix = numpy.zeros((state_space, action_space, state_space))
    reward_noise = numpy.zeros((state_space, action_space, state_space))

    reward_matrix += rewards[None, None, :]
    reward_is_crucial = (numpy.abs(rewards) > 1.0e-6)
    reward_noise_factor = v_scale * random.random()
    reward_noise += reward_noise_factor * (numpy.random.random(size=[state_space]) * reward_is_crucial.astype(float))[None, None, :]
    crucial_index = numpy.where(reward_is_crucial)

    reward_type=comb_sampler(3)
    if(reward_type[0] == '1'): # Add random -s reward
        rm, rn = reward_sampler_sas(state_space, action_space)
        rm[crucial_index] = 0.0 # Make sure the reset triggers are not rewarded
        rn[crucial_index] = 0.0 # Make sure the reset triggers are not rewarded
        reward_matrix += random.exponential(0.10) * reward_scale * rm[None, None, :]
        reward_noise  += random.exponential(0.10) * reward_scale * rn[None, None, :]
    if(reward_type[1] == '1'): # Add random a reward
        reward_matrix += random.exponential(0.01) * reward_scale * (numpy.random.random((1, action_space, 1)) - 0.5)
    if(reward_type[2] == '1'): # Add a random s-s reward
        ndim = task["state_embedding"].shape[1]
        r_direction = numpy.random.random((1, ndim)) # Sample a random reward direction
        r_direction /= numpy.linalg.norm(r_direction)
        r_direction = r_direction * task["state_embedding"]
        r_direction *= random.random() * reward_scale
        r_direction = numpy.sum(r_direction, axis=1)
        reward_matrix += r_direction[None, None, :] - r_direction[:, None, None]

    if(numpy.max(reward_matrix) > 1.01 or numpy.min(reward_matrix) < -1.01):
        reward_noise_type = 'normal'
    else:
        reward_matrix = numpy.clip(reward_matrix, -1.0, 1.0)

    return {"reward": reward_matrix,
            "reward_noise": reward_noise,
            "reward_noise_type": reward_noise_type}

def transition_sampler(state_space:int, 
                    action_space:int, 
                    min_state_space:int, 
                    transition_diversity:int):
        
        # Sample a subset of states
        if(min_state_space is None):
            min_state_space = state_space
        else:
            min_state_space = min(min_state_space, state_space)
        sample_state_space = random.randint(min_state_space, state_space + 1)        
        state_mapping = numpy.random.permutation(state_space)[:sample_state_space]

        if(state_space < 1):
            raise ValueError("State space must be at least 1")

        if(state_space < 2):
            return {"state_mapping": state_mapping,
                    "reset_triggers": numpy.zeros((1,)),
                    "reset_states": numpy.ones((1,)),
                    "transition": numpy.ones((1, action_space, 1)),
                    "reset_triggers_negative": None,
                    "reset_triggers_positive": None,
                    "state_embedding": None}
    
        # Sample the reset states
        eps = 1e-6

        # Now sample dimension of the tasks
        ndim = random.randint(2, 8)

        # Sample the location of states
        loc_s = numpy.random.random((sample_state_space, ndim))

        # Sample the main axis length
        max_axis_length = random.randint(2, sample_state_space)
        loc_s[:, 0] *= max_axis_length

        # Sample positive reset trigger and reset states
        iter = 0
        max_iter = 10000
        need_reset_trigger = (random.random() > 0.70) # 30% probability of not requiring reset triggers
        reset_trigger_positive = numpy.zeros(sample_state_space)
        reset_dist = numpy.zeros(sample_state_space)
        while(((numpy.sum(reset_trigger_positive) < 1 and not need_reset_trigger)
              or numpy.sum(reset_dist) < 1)
              and iter < max_iter):

            proc_s = loc_s[:, 0] / max_axis_length
            max_proc_s = numpy.max(proc_s) - eps
            min_proc_s = numpy.min(proc_s) + eps
            max_proc_s_start = min(0.60, max_proc_s - 0.20)
            min_proc_s_end  = max(0.40, min_proc_s + 0.20)

            if(not need_reset_trigger):
                positive_prob = numpy.clip(proc_s - random.uniform(max_proc_s_start, max_proc_s), 0.0, 1.0)
                positive_prob *= max(1, 0.05 * random.random() * sample_state_space) / (numpy.sum(positive_prob) + eps) 
                positive_prob = numpy.clip(positive_prob, 0.0, 1.0)
                reset_trigger_positive = numpy.random.binomial(1, positive_prob)

            # Sample the reset state distribution
            reset_prob = numpy.clip(random.uniform(min_proc_s, min_proc_s_end) - proc_s, 0.0, 1.0)
            reset_prob *= max(1, 0.05 * random.random() * sample_state_space) / (numpy.sum(reset_prob) + eps) 
            reset_prob = numpy.clip(reset_prob, 0.0, 1.0)
            if(numpy.sum(reset_prob) < 1):
                reset_prob += eps
            reset_dist = numpy.random.binomial(1, reset_prob).astype(float) * (1.0 - reset_trigger_positive) # Avoid resetting to triggers

            # Sample negative reset trigger - Uniform distribution
            # A upperbound of 20% of the state triggers reset
            reset_trigger_negative = numpy.random.binomial(1, 
                                            0.2 * random.random(), 
                                            size=(sample_state_space,)) * \
                                    (1.0 - reset_trigger_positive) * (1.0 - reset_dist)

            reset_trigger = reset_trigger_positive + reset_trigger_negative

            iter += 1

        if(iter >= max_iter):
            raise RuntimeError("Failed to sample a valid task")

        reset_dist = reset_dist / numpy.sum(reset_dist)

        # Sample a upper bound of the transition
        adj_ub = random.randint(2, min(transition_diversity + 1, sample_state_space + 1))

        # Calculate the distance between states
        dist_s = numpy.linalg.norm(loc_s[None, :, :] - loc_s[:, None, :], axis=2)

        # Calculate the minimum k distance between states
        avg_dist = numpy.mean(numpy.sort(dist_s, axis=1)[:, 1:adj_ub])

        # Now sample action of transitions with shape [ns, na, ndim]
        dir_s_a = numpy.random.randn(sample_state_space, action_space, ndim)
        dir_s_a = dir_s_a / numpy.linalg.norm(dir_s_a, axis=2, keepdims=True)
        dist_s_a = (1.0 + 2.0 * numpy.random.random(size=(sample_state_space, action_space))) * avg_dist
        deta_s_a = dir_s_a * dist_s_a[:, :, None]

        # Now calculate the target loc by projecting a with shape[ns, na, ndim]
        target_s_a = deta_s_a + loc_s[:, None, :]

        # Now caculate the state-action-state distance after taking action [ns, na, ns]
        dist_s_a_s = numpy.linalg.norm(target_s_a[:, :, None, :] - loc_s[None, None, :, :], axis=3)
        sigma = avg_dist * random.exponential(scale=1.0)

        prob_s_a_s = numpy.exp(- (dist_s_a_s / sigma) ** 2)

        # Now sample the transition by masking the state-action-state distance
        dist_s_a_s_index = numpy.argsort(dist_s_a_s, axis=2)[:, :, :adj_ub]

        extra_mask = numpy.random.choice([0, 1], size=dist_s_a_s_index.shape)
        extra_mask_sum = extra_mask.sum(axis=2)
        extra_mask_zero = numpy.where(extra_mask_sum == 0)

        for idx in zip(*extra_mask_zero):
            extra_mask[idx[0], idx[1], numpy.random.randint(adj_ub)] = 1
        dist_s_a_s_index = dist_s_a_s_index * extra_mask - (1 - extra_mask)

        # Set the distance beyond n_s_a to inf
        bool_arr = numpy.zeros_like(dist_s_a_s, dtype=bool)
        for i in range(adj_ub):
            bool_arr |= (dist_s_a_s_index[:, :, i][:, :, None] == numpy.arange(sample_state_space)[None, None, :])
        
        prob_s_a_s[bool_arr==False] = 0.0
        transition_matrix = prob_s_a_s / numpy.sum(prob_s_a_s, axis=2, keepdims=True)

        return {"state_mapping": state_mapping,
                "reset_triggers": reset_trigger,
                "reset_states": reset_dist,
                "transition": transition_matrix,
                "reset_triggers_negative": reset_trigger_negative,
                "reset_triggers_positive": reset_trigger_positive,
                "state_embedding": loc_s}

def AnyMDPTaskSampler(state_space:int=128,
                 action_space:int=5,
                 min_state_space:int=None,
                 epoch_state_visit:int=4,
                 max_transition_diversity:int=8,
                 max_iteration:int=50,
                 quality_threshold_transition:float=0.55,
                 quality_threshold_valuefunction:float=0.0,
                 reward_noise_choice:list=['normal'],
                 seed=None,
                 verbose=False):
    # Sampling Transition Matrix and Reward Matrix based on Irwin-Hall Distribution and Gaussian Distribution
    if(seed is not None):
        random.seed(seed)
    else:
        random.seed(pseudo_random_seed())
    
    max_steps = max(int(numpy.random.exponential(epoch_state_visit * state_space)), epoch_state_visit * state_space)

    # Generate Transition Matrix While Check its Quality
    task = {"state_space": state_space,
            "action_space": action_space,
            "max_steps": max_steps}

    qtrans = -1
    qvf = -1

    trans_step = 0
    vf_step = 0

    qtrans_max = -100
    qvf_max = -100
    best_task = None

    while (qtrans < quality_threshold_transition) and trans_step < max_iteration:
        task.update(transition_sampler(state_space, 
                                       action_space,
                                       min_state_space,
                                       max_transition_diversity))
        qtrans = check_task_trans(task)
        if(qtrans > qtrans_max):
            qtrans_max = qtrans
            best_task = deepcopy(task)
        trans_step += 1
    task = best_task

    while (qvf < quality_threshold_valuefunction) and vf_step < max_iteration:
        substate_space = task["state_mapping"].shape[0]
        task.update(reward_sampler(task,
                                   substate_space, 
                                   action_space,
                                   reward_noise_choice))
        qvf = check_task_rewards(task)
        if(qvf > qvf_max):
            qvf_max = qvf
            best_task = deepcopy(task)
        vf_step += 1

    if(verbose):
        print(f"Resample transitions {trans_step} times, quality: {qtrans_max:.2f}",
              f"Rsample rewards {vf_step} times, quality: {qvf_max:.2f}")

    del task["state_embedding"]
    del task["reset_triggers_positive"]
    del task["reset_triggers_negative"]

    return best_task


import numpy
import math
import pygame
from .agent_base import AgentBase
from queue import Queue
from pygame import font
from l3c.mazeworld.envs.dynamics import PI
from l3c.mazeworld.envs.utils import conv2d_numpy
from l3c.mazeworld.envs.dynamics import search_optimal_action
from l3c.mazeworld.envs.ray_caster_utils import landmarks_color, landmarks_rgb, landmarks_rgb_arr, paint_agent_arrow

def convolve_exploration(exp_wht):
    """
    Add weight to the cells where its surroundings are not exposed (mask_info = 0)
    """
    kernel = numpy.ones((5,5))
    kernel[2, 2] = 1000
    res = conv2d_numpy(exp_wht, kernel, padding=2)
    return res

class SmartSLAMAgent(AgentBase):
    def render_init(self, view_size=(480, 480)):
        """
        Initialize a God View With Landmarks
        """
        font.init()
        self._font = font.SysFont("Arial", 18)

        #Initialize the agent drawing
        self._render_cell_size_x = view_size[0] / self._nx
        self._render_cell_size_y = view_size[1] / self._ny
        self._view_size = view_size
        self._window_size = (view_size[0] * 2, view_size[1])

        self._pos_conversion_x = self._render_cell_size_x / self.maze_env.maze_core._cell_size
        self._pos_conversion_y = self._render_cell_size_y / self.maze_env.maze_core._cell_size

        self._screen = pygame.Surface(self._window_size)
        self._screen = pygame.display.set_mode(self._window_size)
        pygame.display.set_caption("AgentRender")
        self._surf_god = pygame.Surface(view_size)
        self._surf_god.fill(pygame.Color("white"))

        for x in range(self._nx):
            for y in range(self._ny):
                if(self._god_info[x,y] < 0):
                    pygame.draw.rect(self._surf_god, pygame.Color("black"), (x * self._render_cell_size_x, y * self._render_cell_size_y,
                            self._render_cell_size_x, self._render_cell_size_y), width=0)

    def render_update(self, observation):
        # paint landmarks
        self._screen.blit(self._surf_god, (0, 0))
        for landmarks_id, (x,y) in enumerate(self._landmarks_coordinates):
            pygame.draw.rect(self._screen, landmarks_color(landmarks_id), 
                    (x * self._render_cell_size_x, y * self._render_cell_size_y,
                    self._render_cell_size_x, self._render_cell_size_y), width=0)

        # paint masks (mists)
        for x in range(self._nx):
            for y in range(self._ny):
                if(self._mask_info[x,y] < 1):
                    pygame.draw.rect(self._screen, pygame.Color("grey"), (x * self._render_cell_size_x, y * self._render_cell_size_y,
                            self._render_cell_size_x, self._render_cell_size_y), width=0)

        # paint agents
        agent_pos = [self._agent_loc[0] * self._pos_conversion_x, self._agent_loc[1] * self._pos_conversion_y]
        paint_agent_arrow(self._screen, pygame.Color("gray"), (0, 0), (agent_pos[0], agent_pos[1]), self._agent_ori, 
                0.4 * self._pos_conversion_x, 0.5 * self._pos_conversion_x)

        # paint target trajectory
        for i in range(0, len(self._path) - 1):
            factor = (i + 1) / len(self._path)
            p = self._path[i]
            n = self._path[i+1]
            p = [(p[0] + 0.5) * self._render_cell_size_x, (p[1] + 0.5) *  self._render_cell_size_y]
            n = [(n[0] + 0.5) * self._render_cell_size_x, (n[1] + 0.5) *  self._render_cell_size_y]
            pygame.draw.line(self._screen, pygame.Color(int(255 * factor), int(255 * (1 - factor)), 128, 255), p, n, width=1)

        # paint the first segment of the path
        p = self._path[0]
        n = self._cur_grid_float
        p = [int((p[0] + 0.5) * self._render_cell_size_x), int((p[1] + 0.5) *  self._render_cell_size_y)]
        n = [int(n[0] * self._render_cell_size_x), int(n[1] *  self._render_cell_size_y)]
        pygame.draw.line(self._screen, pygame.Color(0, 255, 128, 255), p, n, width=1)

        # paint observation
        obs_surf = pygame.surfarray.make_surface(observation)
        obs_surf = pygame.transform.scale(obs_surf, self._view_size)
        self._screen.blit(obs_surf, (self._view_size[0], 0))

        # display
        pygame.display.update()
        done = False
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                done=True
        keys = pygame.key.get_pressed()

    def update_cost_map(self, r_exp=0.25):
        # Calculate Shortest Distance using A*
        # In survival mode, consider the loss brought by rewards
        self._cost_map = 1e+6 * numpy.ones_like(self._god_info)
        refresh_list = Queue()
        cx,cy = self._cur_grid
        for dx,dy in self.valid_neighbors(center=(cx, cy), self_included=True, mask_included=False):
            i = dx + cx
            j = dy + cy
            # Initialize the neiboring costs
            deta = numpy.asarray([(i + 0.5) - self._cur_grid_float[0], (j + 0.5) - self._cur_grid_float[1]])
            dist = numpy.sqrt(numpy.sum(deta ** 2))
            ori = 1.0 - numpy.sum(deta / (dist + 1.0e-3) * numpy.array([numpy.cos(self._agent_ori), numpy.sin(self._agent_ori)]))
            ori_cost = 20.0 * ori * min(dist, 0.01)
            self._cost_map[i, j] = dist + ori_cost
            refresh_list.put((i, j))

        while not refresh_list.empty():
            o_x, o_y = refresh_list.get()
            for d_x, d_y in self.valid_neighbors(center=(o_x, o_y), self_included=False, mask_included=True):
                n_x = o_x + d_x
                n_y = o_y + d_y
                if(n_x >= self._nx or n_x < 0 or n_y >= self._ny or n_y < 0):
                    continue
                c_type = self._god_info[n_x, n_y]
                m_type = self._mask_info[n_x, n_y]
                dist_cost = numpy.sqrt(d_x ** 2 + d_y ** 2)
                if(c_type < 0 and m_type > 0):
                    continue
                elif(m_type < 1):
                    cost = 10 + dist_cost
                else:
                    cost = dist_cost
                if(self._cost_map[n_x, n_y] > self._cost_map[o_x, o_y] + cost):
                    self._cost_map[n_x, n_y] = self._cost_map[o_x, o_y] + cost
                    refresh_list.put((n_x, n_y))

    def policy(self, observation, r):
        self.update_cost_map()
        path_greedy = self.navigate_landmarks_navigate(self._command)
        path = path_greedy
        if(path_greedy is None):
            path_exp = self.exploration()
            if(path_exp is not None):
                path = path_exp
            else:
                print("[WARNING] Unexpected failure in exploration, might cause unexpected stop")
                path = [self._cur_grid]
        self._path = path
        return self.path_to_action(path)

    def path_to_action(self, path):
        if(len(path) < 0):
            raise Exception("Unexpected Error in retrieving path for the current agent")
        d_x = path[0][0] + 0.5 - self._cur_grid_float[0]
        d_y = path[0][1] + 0.5 - self._cur_grid_float[1]
        if(len(path) > 1):
            d_x2 = path[1][0] + 0.5 - self._cur_grid_float[0]
            d_y2 = path[1][1] + 0.5 - self._cur_grid_float[1]
            targ2 = (d_x2, d_y2)

        else:
            targ2 = None
        return search_optimal_action(self._agent_ori, (d_x, d_y), targ2, self._action_space, 1.0)

    def retrieve_path(self, cost_map, goal_idx):
        path = [(int(goal_idx[0]), int(goal_idx[1]))]
        cost = cost_map[goal_idx]
        sel_x, sel_y = goal_idx
        iteration = 0
        eff_targets = []
        for dx, dy in self.valid_neighbors(self_included=True, mask_included=False):
            eff_targets.append((self._cur_grid[0] + dx, self._cur_grid[1] + dy))
        while sel_x != self._cur_grid[0] or sel_y != self._cur_grid[1]:
            flag = False
            for t_x, t_y in eff_targets:
                if(sel_x == t_x and sel_y == t_y):
                    flag = True
            if(flag):
                break
            iteration += 1
            min_cost = cost
            min_x = -1
            min_y = -1
            for d_x, d_y in self.valid_neighbors(center=(sel_x, sel_y)):
                n_x = sel_x + d_x
                n_y = sel_y + d_y
                if(n_x < 0 or n_x > self._nx - 1 or n_y < 0 or n_y > self._ny - 1):
                    continue
                if(cost_map[n_x, n_y] > 1e+4):
                    continue
                if(cost_map[n_x, n_y] < min_cost):
                    # Check whether the location is in shortest path
                    # if not, directly link the agent to the target cell
                    min_cost = cost_map[n_x, n_y]
                    min_x = int(n_x)
                    min_y = int(n_y)
            if(min_x > -1): 
                sel_x, sel_y = (min_x, min_y)
                path.insert(0, (sel_x, sel_y))
                cost=cost_map[sel_x, sel_y]
            else:
                print("[WARNING] Unexpected error in path retrieving")
                break

        if(len(path) > 2):
            d_x = path[0][0] + 0.5 - self._cur_grid_float[0]
            d_y = path[0][1] + 0.5 - self._cur_grid_float[1]
            deta_s = numpy.sqrt(d_x ** 2 + d_y ** 2)
            d_x2 = path[1][0] + 0.5 - self._cur_grid_float[0]
            d_y2 = path[1][1] + 0.5 - self._cur_grid_float[1]
            deta_s2 = numpy.sqrt(d_x2 ** 2 + d_y2 ** 2)
            if(deta_s + cost_map[path[0][0], path[0][1]] > deta_s2 + cost_map[path[1][0], path[1][1]] and deta_s < 0.2):
                del path[0]
        return path

    def exploration(self):
        explore_wht = 1 - numpy.array(self._mask_info, dtype=numpy.int32)
        explore_wht = convolve_exploration(explore_wht)
        utility = self._cost_map - explore_wht
        if(numpy.min(utility) >= 0):
            return None 
        target_idx = numpy.unravel_index(numpy.argmin(utility), utility.shape)
        return self.retrieve_path(self._cost_map, target_idx)

    def navigate_landmarks_navigate(self, landmarks_id):
        idxes = numpy.argwhere(self._god_info == landmarks_id + 1)
        for idx in idxes:
            if(self._mask_info[idx[0], idx[1]] < 1):
                continue
            else:
                return self.retrieve_path(self._cost_map, tuple(idx))
        return None



import sys
import numpy
from queue import Queue
from copy import deepcopy
from l3c.mazeworld.envs.dynamics import PI, DEFAULT_ACTION_SPACE_16, DEFAULT_ACTION_SPACE_32
from l3c.mazeworld.envs.maze_env import MazeWorldContinuous3D


class AgentBase(object):
    """
    Base class for agents
    Use this as parent to create new rule based agents
    """
    def __init__(self, **kwargs):
        self.render = False
        for k in kwargs:
            self.__dict__[k] = kwargs[k]
        if("maze_env" not in kwargs):
            raise Exception("Must use maze_env as arguments")

        # Initialize information
        self._cell_size = self.maze_env.maze_core._cell_size
        self._god_info = 1 - self.maze_env.maze_core._cell_walls + self.maze_env.maze_core._cell_landmarks
        self._landmarks_coordinates = self.maze_env.maze_core._landmarks_coordinates
        self._step_reward = self.maze_env.maze_core._step_reward
        self._nx, self._ny = self._god_info.shape
        self.neighbors = [(-1, 0), (1, 0), (0, 1), (0, -1), (-1, -1), (-1, 1), (1, -1), (1, 1)]
        self._landmarks_visit = dict()
        self._short_term_memory = list()
        self._action_space = self.maze_env.list_actions

        if(self._action_space is None):
            raise Exception("For smart agents, maze environment must use Discrete16 or Discrete32")

        if("short_term_memory_size" not in kwargs):
            self.short_term_memory_size = 3
        if("memory_keep_ratio" not in kwargs):
            self.memory_keep_ratio = 1.0
        self._long_term_memory = numpy.zeros((self._nx, self._ny), dtype=numpy.int8)

        # Render
        if(self.render):
            self.render_init()

    def render_init(self):
        raise NotImplementedError()

    def valid_neighbors(self, center=None, self_included=False, mask_included=True):
        if(center is None):
            cx, cy = self.maze_env.maze_core._agent_grid
        else:
            cx, cy = center
        valid_neighbors = []
        if(self_included):
            valid_neighbors.append((0, 0))
        for dx, dy in self.neighbors:
            nx = cx + dx
            ny = cy + dy
            if(nx < 0 or nx >= self._nx or ny < 0 or ny >= self._ny):
                continue
            if(not self._mask_info[nx, ny] and not mask_included):
                continue
            if(self._god_info[nx, ny] < 0 and self._mask_info[nx, ny]):
                continue
            if(dx * dy == 0):
                valid_neighbors.append((dx, dy))
            else:
                if(self._god_info[nx, cy] > -1 and self._god_info[cx, ny] > -1 
                   and self._mask_info[nx, cy] and self._mask_info[cx, ny]):
                    valid_neighbors.append((dx, dy))
        return valid_neighbors

    def update_common_info(self):
        self._command = self.maze_env.maze_core._command

        # Update long and short term memory
        # Pop the eldest memory from short term memory and insert it to long term memory, but with losses.
        self._short_term_memory.append(numpy.copy(self.maze_env.maze_core._cell_exposed))
        if(len(self._short_term_memory) > self.short_term_memory_size):
            to_longterm = self._short_term_memory.pop(0)
            long_term_keep = (numpy.random.rand(self._nx, self._ny) < self.memory_keep_ratio).astype(numpy.int8)
            self._long_term_memory = numpy.logical_or(self._long_term_memory, to_longterm * long_term_keep)

        # Calculate the current memory: include the long term and short term memory
        self._mask_info = numpy.copy(self._long_term_memory)
        for i in range(len(self._short_term_memory)):
            self._mask_info = numpy.logical_or(self._mask_info, self._short_term_memory[i])
        
        self._agent_ori = self.maze_env.maze_core._agent_ori
        self._agent_loc = self.maze_env.maze_core._agent_loc
        self._cur_grid = deepcopy(self.maze_env.maze_core._agent_grid)
        self._cur_grid_float = deepcopy(self.maze_env.maze_core.get_loc_grid_float(self.maze_env.maze_core._agent_loc))
        lid = self._god_info[self._cur_grid[0], self._cur_grid[1]]
        if(lid > 0):
            self._landmarks_visit[lid - 1] = 0

    def policy(self, observation, r):
        raise NotImplementedError()

    def render_update(self, observation):
        raise NotImplementedError()

    def step(self, observation, r):
        self.update_common_info()
        action = self.policy(observation, r)
        if(self.render):
            self.render_update(observation)
        return action



import numpy as np
from numba import njit

#@njit(cache=True)
def conv2d_numpy(input_data:np.ndarray, 
                 kernel:np.ndarray, 
                 stride=(1,1), padding=0):
    input_height, input_width = input_data.shape
    kernel_height, kernel_width = kernel.shape
    output_height = (input_height - kernel_height + 2 * padding) // stride[0] + 1
    output_width = (input_width - kernel_width + 2 * padding) // stride[1] + 1
    
    output_data = np.zeros((output_height, output_width))
    ni = 0
    for i in range(-padding, input_height - kernel_height + padding + 1, stride[0]):
        ib = max(0, i)
        ie = min(input_height, i + kernel_height)
        _ib = ib - i
        _ie = ie - i
        nj = 0
        for j in range(-padding, input_width - kernel_width + padding + 1, stride[1]):
            jb = max(0, j)
            je = min(input_width, j + kernel_width)
            _jb = jb - j
            _je = je - j
            output_data[ni, nj] = np.sum(input_data[ib:ie, jb:je] * kernel[_ib:_ie, _jb:_je])
            nj += 1
        ni += 1
    
    return output_data
